import numpy 
from . import DEFAULT_RTOL, DEFAULT_ATOL

class Comparison(object):
    '''

    Create an object for storing a comparison expression.

    A comparison expression comprises a relation (such as 'less than')
    and a value (any object capable of being related to with the
    standard comparison operators). These are stored in the
    :attr:`relation` and :attr:`value` attributes respectively.

    The comparison is evaluated for an arbitrary object with the
    :meth:`~cf.Comparison.evaluate` method.

    Valid relations are:

    ========  =============================================
    Relation  Description
    ========  =============================================
    lt        Strictly less than
    le        Less than or equal
    gt        Strictly greater than
    ge        Greater than or equal
    eq        Equals (within a tolerance)
    ne        Not equal to (within a tolerance)
    inside    Inside a given range (range bounds included)
    outside   Outside a given range (range bounds excluded)
    ========  =============================================
 
    .. seealso:: As a convenience, for each relation in the above list
                 there is an identically named function which returns
                 the appropriate :class:`~cf.Comparison` object.

    :keyword **kwargs: A keyword argument to the call is the relation
                       of the comparison (any one of the relations
                       listed above). The keyword's argument is the
                       object to test the comparison against. If the
                       keyword argument is "inside" or "outside" then
                       the value should be a two-element list or
                       tuple.

    ================ =======================================================
    Attribute         (`type`) Description
    ================ =======================================================
    :attr:`relation` (:class:`str`) A string definging the comparison
                     operator. One of relations given above.

    :attr:`value`    (:class:`object`) The object with which to compare.

    ================ =======================================================

    =============================== ====================================================
    Method                          Description
    =============================== ====================================================
    :meth:`~cf.Comparison.dump`     Return a string containing a full description of the
                                    object.

    :meth:`~cf.Comparison.evaluate` Evaluate the comparison for a given variable.
    =============================== ====================================================

    **Examples**

    >>> c = cf.Comparison(le = 5)
    >>> c.evaluate(4)
    True
    >>> c.evaluate(5)
    True

    >>> c = lt(5)
    >>> c.evaluate(4)
    True

    >>> c = cf.Comparison(inside = (1,2))
    >>> a = numpy.arange(4)
    >>> c.evaluate(a)
    array([False,  True,  True, False], dtype=bool)
    >>> (a >= 1) & (a <= 2)
    array([False,  True,  True, False], dtype=bool)

    '''

    def __init__(self, **kwargs):

        for key, value in kwargs.items():
            self.relation = key
            self.value    = value
    #--- End: def Comparison.__init__

    def __repr__(self):

        return '<CF '+str(self)+'>'
    #--- End: def Comparison.__repr__

    def __str__(self):

        return 'Comparison: x '+self.relation+' '+repr(self.value)
    #--- End: def Comparison.__str__

    def dump(self, id=None, nc=False, omit=()):
        
        return str(self)
    #--- End: def Comparison.dump

    def evaluate(self, x):
        '''

    Evaluate the comparison for a given object.

    :parameter x: An object.

    :type x: object

    :returns: Typically a boolean or array of booleans. More
              generally, the output from the appropriate relation
              operators (``==, !=, <, >, <=, >=``) for the given
              variable, x.
        
        '''
        if self.relation == 'lt':
            return x <  self.value
        elif self.relation == 'le':
            return x <= self.value
        elif self.relation == 'gt':
            return x >  self.value
        elif self.relation == 'ge':
            return x >= self.value
        elif self.relation == 'eq':
            return x == self.value
        elif self.relation == 'ne':
            return x != self.value
        elif self.relation == 'inside':
            if self.value[0] < self.value[1]:
                return (x >= self.value[0]) & (x <= self.value[1])
            else:
                return (x >= self.value[1]) & (x <= self.value[0])    
        elif self.relation == 'outside':
            if self.value[0] < self.value[1]:
                return (x <  self.value[0]) | (x >  self.value[1])
            else:
                return (x <  self.value[1]) | (x <  self.value[0])    
    #--- End: def evaluate

#--- End: class Comparison


def lt(value):
    '''
    
    Create a :class:`~cf.Comparison` object for evaluating whether a
    variable is less than the given value.

    :parameter value: The :attr:`value` attribute in the returned
                      :class:`~cf.Comparison` object.

    :type value: object

    :returns: A :class:`~cf.Comparsion` object whose :attr:`relation` attribute is
              set to the string "lt" and whose :attr:`value` attribute set
              to the :attr:`value` parameter.

    **Examples**

    >>> c = cf.lt(5)
    >>> c
    <CF Comparison: x lt 5>
    >>> c.evaluate(4)
    True
    >>> c.evaluate(5)
    False
    '''
    return Comparison(lt=value)
    
def le(value):
    '''
    
    Create a :class:`~cf.Comparsion` object for evaluating whether a variable
    is less than or equal to the given value.

    :parameter value: The :attr:`value` attribute in the returned
                      :class:`~cf.Comparsion` object.

    :type value: object

    :returns: A :class:`~cf.Comparsion` object whose :attr:`relation` attribute is
              set to the string "le" and whose :attr:`value` attribute set
              to the :attr:`value` parameter.

    **Examples**

    >>> c = cf.le(5)
    >>> c
    <CF Comparison: x le 5>
    >>> c.evaluate(5)
    True
    >>> c.evaluate(6)
    False
    '''
    return Comparison(le=value)
    
def gt(value):
    '''
      
    Create a :class:`~cf.Comparsion` object for evaluating whether a variable
    is greater than the given value.

    :parameter value: The :attr:`value` attribute in the returned
                      :class:`~cf.Comparsion` object.

    :type value: object

    :returns: A :class:`~cf.Comparsion` object whose:attr:
              :attr:`relation` attribute is set to the string "gt" and
              whose :attr:`value` attribute set to the :attr:`value`
              parameter.

    **Examples**

    >>> c = cf.gt(5)
    >>> c
    <CF Comparison: x gt 5>
    >>> c.evaluate(6)
    True
    >>> c.evaluate(5)
    False
    '''
    return Comparison(gt=value)
    
def ge(value):
    '''
     
    Create a :class:`~cf.Comparsion` object for evaluating whether a variable
    is greater than or equal to the given value.

    :parameter value: The :attr:`value` attribute in the returned
                      :class:`~cf.Comparsion` object.

    :type value: object

    :returns: A :class:`~cf.Comparsion` object whose :attr:`relation` attribute is
              set to the string "ge" and whose :attr:`value` attribute set
              to the :attr:`value` parameter.

    **Examples**

    >>> c = cf.ge(5)
    >>> c
    <CF Comparison: x ge 5>
    >>> c.evaluate(5)
    True
    >>> c.evaluate(4)
    False
    '''
    return Comparison(ge=value)
    
def eq(value):
    '''
    
    Create a :class:`~cf.Comparsion` object for evaluating whether a variable
    is equal to the given value.

    :parameter value: The :attr:`value` attribute in the returned
                      :class:`~cf.Comparsion` object.

    :type value: object

    :returns: A :class:`~cf.Comparsion` object whose :attr:`relation` attribute is
              set to the string "eq" and whose :attr:`value` attribute set
              to the :attr:`value` parameter.

    **Examples**

    >>> c = cf.eq(5)
    >>> c
    <CF Comparison: x eq 5>
    >>> c.evaluate(5)
    True
    >>> c.evaluate(4)
    False
    '''
    return Comparison(eq=value)
    
def ne(value):
    '''
    
    Create a :class:`~cf.Comparsion` object for evaluating whether a variable
    is not equal to the given value.

    :parameter value: The :attr:`value` attribute in the returned
                      :class:`~cf.Comparsion` object.

    :type value: object

    :returns: A :class:`~cf.Comparsion` object whose :attr:`relation` attribute is
              set to the string "ne" and whose :attr:`value` attribute set
              to the :attr:`value` parameter.

    **Examples**

    >>> c = cf.ne(5)
    >>> c
    <CF Comparison: x ne 5>
    >>> c.evaluate(4)
    True
    >>> c.evaluate(5)
    False
    '''
    return Comparison(ne=value)
    
def inside(value0, value1):
    '''
    
    Create a :class:`~cf.Comparsion` object for evaluating whether a variable
    is inside the given range.

    :parameter value0: The first element of :attr:`value` attribute tuple in
                       the returned :class:`~cf.Comparsion` object.

    :type value0: object

    :parameter value1: The second element of :attr:`value` attribute tuple
                       in the returned :class:`~cf.Comparsion` object.

    :type value1: object

    :returns: A :class:`~cf.Comparsion` object whose :attr:`relation` attribute is
              set to the string "inside" and whose :attr:`value` attribute
              set to the tuple (`value0`, `value1`).

    **Examples**

    >>> c = cf.inside(5,7)
    >>> c
    <CF Comparison: x inside (5, 7)>
    >>> c.evaluate(6)
    True
    >>> c.evaluate(4)
    False

    '''

    return Comparison(inside=(value0, value1))
    
def outside(value0, value1):
    '''
    
    Create a :class:`~cf.Comparsion` object for evaluating whether a
    variable is outside the given range.

    :parameter value0: The first element of :attr:`value` attribute
                       tuple in the returned :class:`~cf.Comparsion`
                       object.

    :type value0: object

    :parameter value1: The second element of :attr:`value` attribute
                       tuple in the returned :class:`~cf.Comparsion`
                       object.

    :type value1: object

    :returns: A :class:`~cf.Comparsion` object whose :attr:`relation`
              attribute is set to the string "outside" and whose
              :attr:`value` attribute set to the tuple (`value0`,
              `value1`).

    **Examples**

    >>> c = cf.outside(5)
    >>> c
    <CF Comparison: x outside (5, 7)>
    >>> c.evaluate(4)
    True
    >>> c.evaluate(6)
    False
    '''
    return Comparison(outside=(value0, value1))
    
def _set_tolerances(x, y, rtol=None, atol=None, override=True):
    '''
    '''
    out = []

    default = DEFAULT_RTOL
    for arg, tol in zip((  rtol ,   atol ),
                        ('_rtol', '_atol')):

        if arg is not None:
            if override:
                # Use the given argument
                pass
            else:
                if hasattr(x, tol) and getattr(x, tol) is not None:
                    arg = getattr(x, tol)                  
                elif hasattr(y, tol) and getattr(y, tol) is not None:
                    arg = getattr(y, tol)                    
                else:
                    # Use the given argument
                    pass
        else:
            # The given argument is None            
            if hasattr(x, tol) and getattr(x, tol) is not None:
                arg = getattr(x, tol)                   
            elif hasattr(y, tol) and getattr(y, tol) is not None:
                arg = getattr(y, tol)
            else:
                arg = default
                
        try:
            arg.__iter__
            arg = default
        except AttributeError:
            pass

        out.append(arg)

        default = DEFAULT_ATOL

    #--- End: for arg, tol

    return tuple(out)
#--- End: def _set_tolerances

def equals(x, y, rtol=None, atol=None, override=True):
    '''Ascertain if x equals y, assuming that they're 1) space
    components, 2) numbers or 3) anything else, in that order.

    dch
    '''

    try: 
        if not x.equals(y, rtol=rtol, atol=atol, override=override):
            return False
    except (AttributeError, TypeError):        
        try:
            if not y.equals(x, rtol=rtol, atol=atol, override=override):
                return False
        except (AttributeError, TypeError):
            # Set rtol and atol
            rtol, atol = _set_tolerances(x, y, rtol=rtol, atol=atol,
                                         override=override)
            try:
                if not numpy.allclose(x, y, rtol=rtol, atol=atol):
                    return False
            except (TypeError, IndexError, NotImplementedError):
                if x != y:
                    return False  

    return True
#--- End: def equals

def dump(x, id=None, nc=False, omit=()):
    '''
    
    Print the full description of an object using its :meth:`dump`
    method.
    
    This function is exactly equivalent to:

    >>> print x.dump(...)

    or, if the object has no :meth:`dump` method:

    >>> print x

    The parameters are only used if the object's :meth:`dump` method
    is being utilized.

    :parameter x: `Optional`. The object to print.

    :type x: object

    :parameter id: `Optional`. Set the common prefix of object's
                   component names, as for the object's :meth:`dump`
                   method.

    :type id: str

    :parameter nc: `Optional`. If True then include attributes whose
                   names begin 'nc'.

    :type nc: bool

    :parameter omit: `Optional`. Omit the given attributes from a
                     objects's description.

    :type omit: sequence
  

    :returns: None

        '''
    try:
        print x.dump(id=id, nc=nc, omit=omit)
    except AttributeError:
        print x
#--- End: def dump
        
